import re
import time
from db.rmdb_client import Client
from decimal import Decimal
from queue import Queue
from threading import Thread
from db.table_layouts import *
from mysql.sql import *
from util import *
from record.record import *

class Driver:
    def __init__(self, scale):
        self._scale = scale
        self._client = Client()
        self._flag = True
        self._delivery_q = Queue()
        self._delivery_t = Thread(target=self.process_delivery, args=(self._delivery_q,))
        self._delivery_t.start()
        self._delivery_stop = False

    def __del__(self):
        self._flag = False
        while not self._delivery_stop:
            continue
        self._client.close()

    def build(self):
        print("Build table schema...")
        sql = open("RDSTester/db/create_tables.sql", "r").read().split(';')
        for line in sql:
            if (line):
                self._client.send_cmd(line + ';')

    def load(self):
        print("Load table data...")
        sql = open("RDSTester/db/load_csvs.sql", "r").read().split(';')
        for line in sql:
            if (line):
                self._client.send_cmd(line + ';')
        print('Database has been initialized.')             
        

    def do_new_order(self, w_id, d_id, c_id, ol_i_id, ol_supply_w_id, ol_quantity):
        ol_cnt = len(ol_i_id)
        ol_amount = 0
        total_amount = 0
        brand_generic = ''

        #transcation
        self._client.send_cmd("BEGIN;")
        print('+ New Order')
        #phase 1
        w_tax = select(client=self._client,
                       table=WAREHOUSE,
                       where=(W_ID,eq,w_id),
                       col=W_TAX)[0][0]
        w_tax = eval(w_tax)
        d_tax, d_next_o_id = select(client=self._client,
                                    table=DISTRICT,
                                    col = (D_TAX,D_NEXT_O_ID),
                                    where=[(D_ID,eq,d_id),
                                           (D_W_ID,eq,w_id)])[0]
        d_tax = eval(d_tax); d_next_o_id = eval(d_next_o_id)
        update(client=self._client,
               table=DISTRICT,
               row=(D_NEXT_O_ID,d_next_o_id+1),
               where=[(D_ID,eq,d_id),(D_W_ID,eq,w_id)])
        try:
            c_discount, c_last_, c_credit = select(client=self._client,
                                               table=CUSTOMER,
                                               col = (C_DISCOUNT,C_LAST,C_CREDIT),
                                               where=[(C_ID,eq,c_id),
                                                      (C_D_ID,eq,d_id),
                                                      (C_W_ID,eq,w_id)])[0]
            c_discount = eval(c_discount)
        except Exception as e:
            c_discount = 0

        #phase 2
        order_time = "'"+current_time()+"'"
        insert(client=self._client,
               table=ORDERS, 
               rows=(d_next_o_id,d_id,w_id,c_id,order_time,0,ol_cnt,int(len(set(ol_supply_w_id))==1)))
        insert(client=self._client,
                table=NEW_ORDERS,
               rows=(d_next_o_id,d_id,w_id))
        #phase 3
        for i in range(ol_cnt):
            i_price, i_name, i_data = select(client=self._client,
                                            table=ITEM,
                                            col=(I_PRICE,I_NAME,I_DATA),
                                            where=(I_ID,eq,ol_i_id[i]))[0]
            try: 
                i_price = eval(i_price)
            except Exception as e:
                i_price = 1
                print("Error:", e)  # 打印异常信息

            s_quantity, *s_dist, s_ytd, s_order_cnt, s_remote_cnt, s_data= \
            select(client=self._client,
                   table=STOCK,
                   col=(S_QUANTITY,S_DIST_01,S_DIST_02,S_DIST_03,S_DIST_04,S_DIST_05,S_DIST_06,S_DIST_07,
                        S_DIST_08,S_DIST_09,S_DIST_10,S_YTD,S_ORDER_CNT,S_REMOTE_CNT,S_DATA),
                   where=[(S_I_ID,eq,ol_i_id[i]),
                          (S_W_ID,eq,ol_supply_w_id[i])])[0]
            try:
                s_quantity = eval(s_quantity); s_ytd = eval(s_ytd); s_order_cnt = eval(s_order_cnt); s_remote_cnt = eval(s_remote_cnt); 
            except Exception as e:
                s_quantity = 0; s_ytd = 0; s_order_cnt = 0; s_remote_cnt = 0; 
                print("Error:", e)  # 打印异常信息
            if s_quantity - ol_quantity[i] >= 10:
                s_quantity -= ol_quantity[i]
            else:
                s_quantity = s_quantity - ol_quantity[i] + 91
            s_ytd += ol_quantity[i]
            s_order_cnt += 1
            if ol_supply_w_id[i] != w_id:
                s_remote_cnt += 1
            update(client=self._client,
                   table=STOCK,
                   row=[(S_QUANTITY,s_quantity),
                        (S_YTD,s_ytd),
                        (S_ORDER_CNT,s_order_cnt),
                        (S_REMOTE_CNT,s_remote_cnt)],
                   where=[(S_I_ID, eq, ol_i_id[i]),
                          (S_W_ID, eq, ol_supply_w_id[i])])
            ol_amount = ol_quantity[i] * i_price
            brand_generic = 'B' if re.search('ORIGINAL',i_data) and re.search('ORIGINAL',s_data) else 'G'
            insert(client=self._client,
                   table=ORDER_LINE,
                   rows=(d_next_o_id,d_id,w_id,i,ol_i_id[i],ol_supply_w_id[i],order_time,ol_quantity[i],ol_amount,"'"+s_dist[d_id-1]+"'"))
            total_amount += ol_amount
        total_amount *=(1-c_discount)*(1+w_tax+d_tax)
        self._client.send_cmd("COMMIT;")
        print('- New Order')
        return True


    def do_payment(self, w_id, d_id, c_w_id, c_d_id, c_query, h_amount):
        self._client.send_cmd("BEGIN;")
        print('+ Payment')
        w_name, w_street_1, w_street_2, w_city, w_state, w_zip, w_ytd = \
        select(client=self._client,
               table=WAREHOUSE,
               col=(W_NAME,W_STREET_1,W_STREET_2,W_CITY,W_STATE,W_ZIP,W_YTD),
               where=(W_ID,eq,w_id))[0]
        w_ytd = eval(w_ytd)
        update(client=self._client,
               table=WAREHOUSE,
               row=(W_YTD,w_ytd+h_amount),
               where=(W_ID,eq,w_id))
        d_name, d_street_1, d_street_2, d_city, d_state, d_zip, d_ytd = \
        select(client=self._client,
               table=DISTRICT,
               col=(D_NAME,D_STREET_1,D_STREET_2,D_CITY,D_STATE, D_ZIP,D_YTD),
               where=(D_ID,eq,d_id))[0]
        d_ytd = eval(d_ytd)
        update(client=self._client,
               table=DISTRICT,
               row=(D_YTD,d_ytd+h_amount),
               where=[(D_W_ID,eq,w_id),(D_ID,eq,d_id)])
        if(type(c_query) == str):
            result = select(client=self._client,
                            table=CUSTOMER,
                            col=(C_ID,C_FIRST,C_MIDDLE,C_LAST,C_STREET_1,C_STREET_2,C_CITY,C_STATE,
                                           C_ZIP,C_PHONE,C_SINCE,C_CREDIT,C_CREDIT_LIM,C_DISCOUNT,
                                           C_BALANCE,C_YTD_PAYMENT,C_PAYMENT_CNT),
                            where=[(C_LAST,eq,c_query),
                                   (C_W_ID,eq,c_w_id),
                                   (C_D_ID,eq,c_d_id)],
                            order_by=C_FIRST,
                            asc=True)[0]
        else:
            try:
                result = select(client=self._client,
                            table=CUSTOMER,
                            col=(C_ID, C_FIRST, C_MIDDLE, C_LAST, C_STREET_1, C_STREET_2, C_CITY, C_STATE,
                                 C_ZIP, C_PHONE, C_SINCE, C_CREDIT, C_CREDIT_LIM, C_DISCOUNT,
                                 C_BALANCE, C_YTD_PAYMENT, C_PAYMENT_CNT),
                            where=[(C_ID, eq, c_query),
                                   (C_W_ID, eq, c_w_id),
                                   (C_D_ID, eq, c_d_id)])[0]
                c_id, c_first, c_midele, c_last, \
                c_street_1,c_street_2,c_city,c_state,\
                c_zip,c_phone,c_since,\
                c_credit,c_credit_lim,c_discount,c_balance,c_ytd_payment,c_payment_cnt = result  # result[len(result)//2]
                c_id = eval(c_id); c_balance = eval(c_balance); c_ytd_payment = eval(c_ytd_payment); c_payment_cnt = eval(c_payment_cnt)
            except Exception as e:
                c_credit = 'GC'
                c_id = 1; c_balance = 0; c_ytd_payment = 0; c_payment_cnt = 0
        update(client=self._client,
               table=CUSTOMER,
               row=[(C_BALANCE,c_balance+h_amount),
                    (C_YTD_PAYMENT,c_ytd_payment+1),
                    (C_PAYMENT_CNT,c_payment_cnt+1)],
               where=[(C_W_ID,eq,w_id),(C_D_ID,eq,d_id),(C_ID,eq,c_id)])
        if c_credit == 'BC':
            c_data = (''.join(map(str,[c_id,c_d_id,c_w_id,d_id,h_amount]))+select(client=self._client,
                                                                                  table=CUSTOMER,
                                                                                  col=C_DATA,
                                                                                  where=[(C_ID,eq,c_id),
                                                                                         (C_W_ID,eq,c_w_id),
                                                                                         (C_D_ID,eq,c_d_id)])[0][0] )[0:500]
            update(client=self._client,
                   table=CUSTOMER,
                   row=(C_DATA,"'"+c_data+"'"),
                   where=[(C_W_ID,eq,w_id),(C_D_ID,eq,d_id),(C_ID,eq,c_id)])
        #4 blank space
        h_data = w_name + '    ' + d_name
        insert(client=self._client,
               table=HISTORY,
               rows=(c_id,c_d_id,c_w_id,d_id,w_id,"'"+current_time()+"'",h_amount,"'"+h_data+"'"))
        self._client.send_cmd("COMMIT;")
        print('- Payment')
        return True

    def do_order_status(self, w_id, d_id, c_query):
        self._client.send_cmd("BEGIN;")
        print('+ Order Status')
        if type(c_query) == str:
            result = select(client=self._client,
                            table=CUSTOMER,
                            col=(C_ID, C_BALANCE, C_FIRST, C_MIDDLE, C_LAST),
                            where=[(C_LAST, eq, c_query),
                                   (C_W_ID, eq, w_id),
                                   (C_D_ID, eq, d_id)],
                            order_by=C_FIRST,
                            asc=True)[0]
        else:
            try:
                result = select(client=self._client,
                            table=CUSTOMER,
                            col=(C_ID, C_BALANCE, C_FIRST, C_MIDDLE, C_LAST),
                            where=[(C_ID, eq, c_query),
                                   (C_W_ID, eq, w_id),
                                   (C_D_ID, eq, d_id)])[0]
                c_id, c_balance, c_first, c_middle, c_last = result  # result[len(result)//2]
                c_id = eval(c_id)
            except Exception as e:
                c_id = 2101
        try:
            o_id, o_entry_id, o_carrier_id = select(client=self._client,
                                                table=ORDERS,
                                                col=(O_ID,O_ENTRY_D,O_CARRIER_ID),
                                                where=[(O_W_ID, eq, w_id),
                                                       (O_D_ID, eq, d_id),
                                                       (O_C_ID, eq, c_id)],
                                                order_by=O_ID)[0]
            o_id = eval(o_id)
        except Exception as e:
            o_id = 1
        try:
            result = select(client=self._client,#ol_i_id,ol_supply_w_id,ol_quantity,ol_amount,ol_delivery_d
                        table=ORDER_LINE,
                        col=(OL_I_ID, OL_SUPPLY_W_ID, OL_QUANTITY,OL_AMOUNT,OL_DELIVERY_D),
                        where=[(OL_W_ID,eq,w_id),
                               (OL_D_ID,eq,d_id),
                               (OL_O_ID,eq,o_id)])[0]
        except Exception as e:
            result = None
        self._client.send_cmd("COMMIT;")
        print('- Order Status')
        return True


    def do_delivery(self, w_id, o_carrier_id):
        self._delivery_q.put({'w_id':w_id,'o_carrier_id':o_carrier_id})

    def process_delivery(self, q):
        tmp_client = Client()
        while self._flag:
            if not q.empty():
                t1 = time.time()
                tmp_client.send_cmd("BEGIN;")
                print('+ Delivery')
                dat = q.get()
                w_id = dat['w_id']
                o_carrier_id = dat['o_carrier_id']
                for d_id in range(1,11):
                    try:
                        o_id = select(client=tmp_client,
                                    table=NEW_ORDERS,
                                    col=NO_O_ID,
                                    where=[(NO_W_ID,eq,w_id),(NO_D_ID,eq,d_id)],
                                    order_by=NO_O_ID,
                                    asc=True)[0][0]
                        o_id = eval(o_id)
                    except Exception as e:
                        o_id = 2101
                    delete(client=tmp_client,
                            table=NEW_ORDERS,
                            where=[(NO_W_ID,eq,w_id),(NO_D_ID,eq,d_id),(NO_O_ID,eq,o_id)])

                    o_c_id = select(client=tmp_client,
                                    table=ORDERS,
                                    col=O_C_ID,
                                    where=[(O_ID,eq,o_id),(O_W_ID,eq,w_id),(O_D_ID,eq,d_id)])[0][0]
                    o_c_id = eval(o_c_id)
                    update(client=tmp_client,
                           table=ORDERS,
                           row=(O_CARRIER_ID,o_carrier_id),
                           where=[(O_ID,eq,o_id),(O_W_ID,eq,w_id),(O_D_ID,eq,d_id)])
                    order_lines = select(client=tmp_client,
                                         table=ORDER_LINE,
                                         where=[(OL_W_ID,eq,w_id),(OL_D_ID,eq,d_id),(OL_O_ID,eq,o_id)])
                    ol_amount = select(client=tmp_client,
                                       table=ORDER_LINE,
                                       #col='sum(' + OL_AMOUNT + ')',
                                       col = OL_AMOUNT,
                                       where=[(OL_W_ID, eq, w_id), (OL_D_ID, eq, d_id), (OL_O_ID, eq, o_id)]) # [0]
                    #print(ol_amount)
                    ol_amount = [eval(o[0]) for o in ol_amount]
                    #print(ol_amount)
                    ol_amount = sum(ol_amount)
                    for line in order_lines:
                        update(client=tmp_client,
                               table=ORDER_LINE,
                               row=(OL_DELIVERY_D,"'"+current_time()+"'"),
                               where=[(OL_W_ID,eq,w_id),(OL_D_ID,eq,d_id),(OL_O_ID,eq,eval(line[0]))])

                    c_balance,c_delivery_cnt = select(client=tmp_client,
                                                      table=CUSTOMER,
                                                      col=(C_BALANCE,C_DELIVERY_CNT),
                                                      where=[(C_W_ID,eq,w_id),(C_D_ID,eq,d_id),(C_ID,eq,o_c_id)])[0]
                    c_balance = eval(c_balance); c_delivery_cnt = eval(c_delivery_cnt)
                    #print(c_balance, ol_amount, c_delivery_cnt)
                    update(client=tmp_client,
                           table=CUSTOMER,
                           row=[(C_BALANCE,c_balance+ol_amount),(C_DELIVERY_CNT,c_delivery_cnt+1)],
                           where=[(C_W_ID,eq,w_id),(C_D_ID,eq,d_id),(C_ID,eq,o_c_id)])
                tmp_client.send_cmd("COMMIT;")
                t2 = time.time()
                put_txn(Delivery,t2-t1,True)
                print('- Delivery')

        tmp_client.close()
        self._delivery_stop = True


    def do_stock_level(self, w_id, d_id, threshold):
        self._client.send_cmd("BEGIN;")
        print('+ Stock Level')
        d_next_o_id = select(client=self._client,
                             table=DISTRICT,
                             col=D_NEXT_O_ID,
                             where=[(D_W_ID,eq,w_id),(D_ID,eq,d_id)])[0][0]
        d_next_o_id = eval(d_next_o_id)
        order_lines = select(client=self._client,
                             table=ORDER_LINE,
                             where=[(OL_W_ID,eq,w_id),
                                    (OL_D_ID,eq,d_id),
                                    (OL_O_ID,beq,d_next_o_id-20),
                                    (OL_O_ID,lt,d_next_o_id)])
        items = set([order_line[4] for order_line in order_lines])
        low_stock = 0
        for item in items:
            cur_quantity = select(client=self._client,
                                table=STOCK,
                                col=S_QUANTITY,
                                where=[(S_I_ID,eq,item),
                                        (S_W_ID,eq,w_id),
                                        (S_QUANTITY,lt,threshold)])[0][0]
            # low_stock += eval(cur_quantity)
        self._client.send_cmd("COMMIT;")
        print('- Stock Level')
        return True